﻿using C1.Win.C1FlexGrid;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

namespace C1FlexGridOutlinesAndTreesUnbound
{
    public partial class Form1 : Form
    {

        //------------------------------------------------------------------------------------

        #region ** fields

        List<object[]> dataRows = new List<object[]>();
        CellStyle[] _nodeStyle = new CellStyle[2];
        CellStyle[] _lockStyle = new CellStyle[2];

        int _lastPreviousRowIndex;
        int _lastNextRowIndex;
        int _movedTopNodeRowIndex;

        #endregion

        //------------------------------------------------------------------------------------

        #region ** ctor

        public Form1()
        {
            InitializeComponent();

            // customize outline styles
            var cs = _nodeStyle[0] = _flex.Styles[CellStyleEnum.Subtotal0];
            cs.BackColor = Color.FromArgb(0xd0, 0xd0, 0xd0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Bold);
            cs = _nodeStyle[1] = _flex.Styles[CellStyleEnum.Subtotal1];
            cs.BackColor = Color.FromArgb(0xe0, 0xe0, 0xe0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Italic);

            // styles for locked nodes
            cs = _lockStyle[0] = _flex.Styles.Add("LockSubtotal0", "Subtotal0");
            cs.ForeColor = Color.Red;
            cs = _lockStyle[1] = _flex.Styles.Add("LockSubtotal1", "Subtotal1");
            cs.ForeColor = Color.Red;

            // remove vertical gridlines from detail area
            _flex.Styles.Normal.Border.Direction = BorderDirEnum.Horizontal;

            using (new DeferRefresh(_flex))
            {
                ResetData();
            }
        }

        #endregion

        //------------------------------------------------------------------------------------

        #region ** low-level methods: grouping, trees, subtotals with InsertNode and Aggregate

        // clear all data from grid and add new
        void ResetData()
        {
            // add columns for data, fixed columns and rows for headers, assign proper DataType
            SetupGrid();

            // prepare data for grid 
            GenerateDataRows();

            // fill grid
            PopulateGrid();

            // resize columns to fit their content
            _flex.AutoSizeCols();

            _lastPreviousRowIndex = -1;
            _lastNextRowIndex = -1;
            _movedTopNodeRowIndex = -1;
        }

        void SetupGrid()
        {
            // clear grid
            _flex.Clear();
            _flex.Cols.Count = _flex.Cols.Fixed = fixedColumnToolStripMenuItem.Checked ? 1 : 0;
            _flex.Rows.Count = _flex.Rows.Fixed = 1;
            _flex.Tree.Column = -1;

            // prepare grid for filling with data
            _flex.Cols.Add(6);

            // start from first non-fixed column for setup grid
            int columnIndex = _flex.Cols.Fixed;
            _flex.Cols[columnIndex].DataType = typeof (string);
            _flex.Cols[columnIndex].Caption = "Country";
            _flex.Cols[columnIndex].Name = "Country";

            columnIndex++;
            _flex.Cols[columnIndex].DataType = typeof (string);
            _flex.Cols[columnIndex].Caption = "City";
            _flex.Cols[columnIndex].Name = "City";

            columnIndex++;
            _flex.Cols[columnIndex].DataType = typeof (string);
            _flex.Cols[columnIndex].Caption = "Product";
            _flex.Cols[columnIndex].Name = "Product";

            columnIndex++;
            _flex.Cols[columnIndex].DataType = typeof (decimal);
            _flex.Cols[columnIndex].Caption = "Price";
            _flex.Cols[columnIndex].Name = "Price";

            columnIndex++;
            _flex.Cols[columnIndex].DataType = typeof (short);
            _flex.Cols[columnIndex].Caption = "Quantity";
            _flex.Cols[columnIndex].Name = "Quantity";

            columnIndex++;
            _flex.Cols[columnIndex].DataType = typeof (decimal);
            _flex.Cols[columnIndex].Caption = "Cost";
            _flex.Cols[columnIndex].Name = "Cost";
            _flex.Cols[columnIndex].AllowEditing = false;

            // format Price column
            _flex.Cols["Price"].Format = "n2";

            // format Cost column
            _flex.Cols["Cost"].Format = "n2";

            // allow node content to spill onto next cell
            _flex.AllowMerging = AllowMergingEnum.Nodes;
        }

        void GenerateDataRows()
        {
            // fill only if data does not exist
            if (dataRows.Count == 0)
            {
                // generate rows with random data for grid
                Random random = new Random();
                int currentRecordIndex = _flex.Rows.Count - _flex.Rows.Fixed;
                int currentCountryIndex = 0;
                int currentCityIndex = 0;

                while (currentRecordIndex < Consts.maxRecordCount &&
                       (currentCountryIndex != Consts.countriesAndCities.Count - 1 ||
                        (currentCountryIndex == Consts.countriesAndCities.Count - 1 &&
                         currentCityIndex != Consts.countriesAndCities.Last().Value.Length - 1)))
                {
                    bool useCurrentCity = random.Next(0, 2) == 1;

                    // use current city
                    if (useCurrentCity)
                    {
                        object[] dataRow = new object[6];
                        dataRow[0] = Consts.countriesAndCities.ElementAt(currentCountryIndex).Key;
                        dataRow[1] = Consts.countriesAndCities.ElementAt(currentCountryIndex).Value[currentCityIndex];
                        dataRow[2] = Consts.products[random.Next(0, Consts.products.Length)];

                        // generate price
                        Decimal price = Convert.ToDecimal(Convert.ToSingle(random.Next(0, 100000))/100);
                        dataRow[3] = price;

                        // generate quantity
                        short quantity = Convert.ToInt16(random.Next(0, 1000));
                        dataRow[4] = quantity;

                        // calculate cost
                        dataRow[5] = Convert.ToDecimal(price*quantity);
                        dataRows.Add(dataRow);
                    }

                    // use next city
                    else if (currentCityIndex < Consts.countriesAndCities.ElementAt(currentCountryIndex).Value.Length - 1)
                        currentCityIndex++;

                    // use next country
                    else
                    {
                        currentCountryIndex++;
                        currentCityIndex = 0;
                    }

                    // move to next row of data
                    currentRecordIndex++;
                }
            }
        }

        void PopulateGrid()
        {
            // populate grid with generated data
            for (int i = 0; i < dataRows.Count; i++)
            {
                // handle fixed columns
                if (dataRows[i].Length < _flex.Cols.Count)
                {
                    object[] tmpRow = dataRows[i];
                    dataRows[i] = new object[_flex.Cols.Count];
                    dataRows[i][0] = null;
                    Array.Copy(tmpRow, 0, dataRows[i], 1, tmpRow.Length);
                }
                if (dataRows[i].Length > _flex.Cols.Count)
                {
                    object[] tmpRow = dataRows[i];
                    dataRows[i] = new object[_flex.Cols.Count];
                    Array.Copy(tmpRow, 1, dataRows[i], 0, dataRows[i].Length);
                }
                _flex.AddItem(dataRows[i]);
            }
        }

        // group on a given column inserting nodes of a given level
        void GroupBy(string columnName, int level)
        {
            object current = null;
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (!_flex.Rows[r].IsNode)
                {
                    var value = _flex[r, columnName];
                    if (!object.Equals(value, current))
                    {
                        // value changed: insert node, apply style
                        _flex.Rows.InsertNode(r, level);
                        _flex.Rows[r].Style = _nodeStyle[Math.Min(level, _nodeStyle.Length - 1)];

                        // show group name in first scrollable column
                        _flex[r, _flex.Cols.Fixed] = value;

                        // update current value
                        current = value;
                    }
                }
            }
        }

        // add subtotals to each node at a given level
        void AddSubtotals(int level, string colName)
        {
            // get column we are going to total on
            int colIndex = _flex.Cols.IndexOf(colName);

            // scan rows looking for nodes at the right level
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (_flex.Rows[r].IsNode)
                {
                    var node = _flex.Rows[r].Node;
                    if (node.Level == level)
                    {
                        // found a node, calculate the sum of cost
                        var range = node.GetCellRange();
                        var sum = _flex.Aggregate(AggregateEnum.Sum,
                            range.r1, colIndex, range.r2, colIndex,
                            AggregateFlags.ExcludeNodes);

                        // show the sum on the grid 
                        // (will use the column format automatically)
                        _flex[r, colIndex] = sum;
                    }
                }
            }
        }

        // change style of subtotals of one top-level node
        void ChangeSubtotalsStyle(CellRange range, CellStyle[] styles)
        {
            for (int i = range.TopRow; i <= range.BottomRow; i++)
                if (_flex.Rows[i].IsNode)
                {
                    _flex.Rows[i].Style = styles[Math.Min(_flex.Rows[i].Node.Level, styles.Length - 1)];
                }
        }

    void MoveSelectedTopNodeToGridTop()
    {
        Node nodeToMove = null;

        if (_flex.Rows[_flex.Row].IsNode)
            nodeToMove = _flex.Rows[_flex.Row].Node;
        else
        {
            // current selected row
            int rowIndex = _flex.Row;

            // search node above row with data or below
            int increment = _flex.SubtotalPosition == SubtotalPositionEnum.AboveData ? -1 : 1;

            // search deepest node
            while (rowIndex > _flex.Rows.Fixed)
            {
                if (_flex.Rows[rowIndex].IsNode)
                {
                    nodeToMove = _flex.Rows[rowIndex].Node;
                    break;
                }
                rowIndex += increment;
            }
        }

        if (nodeToMove != null)
        {
            while (nodeToMove.Level != 0)
                nodeToMove = nodeToMove.Parent;

            // restoring position of previous moved node
            if (_movedTopNodeRowIndex != -1)
            {
                // getting all cells of previous moved top-level node
                CellRange movedTopNodeRange = _flex.Rows[_movedTopNodeRowIndex].Node.GetCellRange();

                // revert subtotals styles
                ChangeSubtotalsStyle(movedTopNodeRange, _nodeStyle);

                // move node with all its children to initial position
                int rowsCount = movedTopNodeRange.BottomRow - movedTopNodeRange.TopRow + 1;
                if (_lastPreviousRowIndex != -1)
                    _flex.Rows.MoveRange(movedTopNodeRange.r1, rowsCount, _lastPreviousRowIndex - rowsCount + 1);
                else if (_lastNextRowIndex != -1)
                    _flex.Rows.MoveRange(movedTopNodeRange.r1, rowsCount, _lastNextRowIndex - rowsCount + 1);
            }

            // find previous and next node positions for reverting position previous moved node
            Node prevNode = nodeToMove.PrevNode;
            Node nextNode = nodeToMove.NextNode;

            // moving node to top of the grid
            nodeToMove.Move(NodeMoveEnum.First);

            // store indexes of previous, next and moved rows
            if (prevNode != null)
            {
                CellRange cr = prevNode.GetCellRange();
                _lastPreviousRowIndex = cr.BottomRow;
            }
            else
                _lastPreviousRowIndex = -1;

            if (nextNode != null)
                _lastNextRowIndex = nextNode.Row.Index - 1;
            else
                _lastNextRowIndex = -1;

            _movedTopNodeRowIndex = nodeToMove.Row.Index;
            _flex.TopRow = nodeToMove.Row.Index;

            // getting all cells of just moved top-level node
            CellRange nodeToMoveRange = nodeToMove.GetCellRange();

            // apply lock subtotals styles
            ChangeSubtotalsStyle(nodeToMoveRange, _lockStyle);
        }
    }

        #endregion

        //------------------------------------------------------------------------------------

        #region ** event handlers

        void _btnReset_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetData();
            }
        }

        void _btnGroupCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetData();
                GroupBy("Country", 0);
            }
        }

        void _btnGroupCountryCity_Click(object sender, EventArgs e)
        {
            // suspend redrawing while updating
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                // group by Country, City
                GroupBy("Country", 0);
                GroupBy("City", 1);
            }
        }

        void _btnTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country as before
                _btnGroupCountry_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(0);
            }
        }

        void _btnTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country and city as before
                _btnGroupCountryCity_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }

        void _btnNiceTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountry_Click(this, EventArgs.Empty);

                // add totals per country
                AddSubtotals(0, "Cost"); // cost per country (level 0)
                AddSubtotals(0, "Quantity"); // quantity per country (level 0)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }

        void _btnNiceTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountryCity_Click(this, EventArgs.Empty);

                // add totals per country, city
                AddSubtotals(0, "Cost"); // cost per country (level 0)
                AddSubtotals(0, "Quantity"); // quantity per country (level 0)
                AddSubtotals(1, "Cost"); // cost per city (level 1)
                AddSubtotals(1, "Quantity"); // quantity per city (level 1)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
            }
        }

        void _btnGroupCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetData();
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
            }
        }

        void _btnGroupCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                // group by Country, City
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "Cost");
            }
        }

        void _btnTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                // group by country as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(0);
            }
        }

        void _btnTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                // group by country and city as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }

        void _btnNiceTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                // group and total by country
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }

        void _btnNiceTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetData();

                _flex.Tree.Column = 0;

                // group and total by country and city
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Quantity");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "Quantity");

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
                _flex.AllowMerging = AllowMergingEnum.Nodes;

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);
            }
        }

        private void moveToTopToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MoveSelectedTopNodeToGridTop();
        }

        void fixedColumnToolStripMenuItem_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                fixedColumnToolStripMenuItem.Checked = !fixedColumnToolStripMenuItem.Checked;
                ResetData();
            }
        }

        private void _flex_CellChanged(object sender, RowColEventArgs e)
        {
            // check in what column data was changed
            if (_flex.Cols[e.Col].Name == "Price" || _flex.Cols[e.Col].Name == "Quantity")
            {
                // check data
                if (Convert.ToInt32(_flex[e.Row, e.Col]) < 0)
                    _flex[e.Row, e.Col] = 0;

                // calculate new cost
                if (_flex[e.Row, "Price"] != null && _flex[e.Row, "Quantity"] != null)
                    _flex[e.Row, "Cost"] = (decimal)_flex[e.Row, "Price"] * (short)_flex[e.Row, "Quantity"];
            }
        }

        #endregion
    }



    /// <summary>
    /// Utility class used to encapsulate grid lengthy operations in BeginUpdate/EndUpdate block.
    /// This avoids flicker and ensures that all updates are done in case 
    /// an exception is thrown during the operation.
    /// </summary>
    class DeferRefresh : IDisposable
    {
        C1FlexGrid _grid;
        public DeferRefresh(C1FlexGrid grid)
        {
            _grid = grid;
            _grid.BeginUpdate();
        }
        public void Dispose()
        {
            _grid.EndUpdate();
        }
    }
}
